from collections import defaultdict
from textwrap import dedent
from uuid import UUID

from django.core.exceptions import ValidationError as DjangoValidationError
from django.db.models import Q
from django.shortcuts import get_object_or_404
from drf_spectacular.utils import OpenApiExample, OpenApiParameter, OpenApiResponse, extend_schema, extend_schema_view
from rest_framework import permissions, serializers, status
from rest_framework.exceptions import NotFound, PermissionDenied, ValidationError
from rest_framework.generics import CreateAPIView, ListAPIView, ListCreateAPIView, RetrieveUpdateDestroyAPIView
from rest_framework.response import Response

from arkindex.documents.models import (
    Corpus,
    Element,
    Entity,
    EntityLink,
    EntityRole,
    EntityType,
    Transcription,
    TranscriptionEntity,
)
from arkindex.documents.serializers.elements import ElementTinySerializer
from arkindex.documents.serializers.entities import (
    BaseEntitySerializer,
    CreateEntityRoleErrorResponseSerializer,
    EntityCreateSerializer,
    EntityLinkCreateSerializer,
    EntityLinkSerializer,
    EntityRoleSerializer,
    EntitySerializer,
    EntityTypeCreateSerializer,
    EntityTypeSerializer,
    TranscriptionEntitiesBulkSerializer,
    TranscriptionEntityCreateSerializer,
    TranscriptionEntitySerializer,
)
from arkindex.documents.serializers.light import EntityTypeLightSerializer
from arkindex.process.models import WorkerRun, WorkerVersion
from arkindex.project.mixins import ACLMixin, CorpusACLMixin
from arkindex.project.openapi import UUID_OR_FALSE
from arkindex.project.permissions import IsVerified, IsVerifiedOrReadOnly
from arkindex.users.models import Role


@extend_schema(tags=["entities"])
@extend_schema_view(
    get=extend_schema(operation_id="ListCorpusRoles", description="List all roles of a corpus"),
    post=extend_schema(
        description="Create a new entity role",
        responses={
            200: EntityRoleSerializer,
            400: CreateEntityRoleErrorResponseSerializer
        },
        examples=[OpenApiExample(
            status_codes=["400"],
            response_only=True,
            name="role-exists",
            value={"id": "55cd009d-cd4b-4ec2-a475-b060f98f9138", "corpus": ["Role already exists in this corpus"]},
            description="Role already exists."
        )]
    )
)
class CorpusRoles(CorpusACLMixin, ListCreateAPIView):
    """
    List all roles in a corpus
    """
    permission_classes = (IsVerifiedOrReadOnly, )
    serializer_class = EntityRoleSerializer
    queryset = EntityRole.objects.none()

    def get_queryset(self):
        return EntityRole.objects \
            .filter(corpus=self.get_corpus(self.kwargs["pk"])) \
            .order_by("parent_name", "child_name")

    def perform_create(self, serializer):
        data = self.request.data
        if EntityRole.objects.filter(
            parent_name=data["parent_name"],
            child_name=data["child_name"],
            parent_type=data["parent_type_id"],
            child_type=data["child_type_id"],
            corpus_id=self.request.parser_context["kwargs"]["pk"]
        ).exists():
            raise serializers.ValidationError({
                "corpus": ["Role already exists in this corpus"],
                "id": self.request.parser_context["kwargs"]["pk"]
            })
        super().perform_create(serializer)


@extend_schema(tags=["entities"])
@extend_schema_view(
    get=extend_schema(operation_id="ListCorpusEntityTypes", description="List all entity types in a corpus"),
)
class CorpusEntityTypes(CorpusACLMixin, ListAPIView):
    permission_classes = (IsVerifiedOrReadOnly, )
    serializer_class = EntityTypeLightSerializer
    queryset = EntityType.objects.none()

    def get_queryset(self):
        return EntityType.objects \
            .filter(corpus=self.get_corpus(self.kwargs["pk"])) \
            .order_by("name")


@extend_schema_view(post=extend_schema(
    operation_id="CreateEntityType",
    responses={201: EntityTypeSerializer},
    tags=["entities"],
))
class EntityTypeCreate(CreateAPIView):
    """
    Create a new Entity Type in a corpus.
    Requires admin access to the corpus.
    """
    serializer_class = EntityTypeCreateSerializer
    permission_classes = (IsVerified, )


@extend_schema_view(
    get=extend_schema(
        operation_id="RetrieveEntityType",
        description=dedent("""
        Retrieve an existing entity type.

        This requires guest access to the entity type's corpus.
        """),
    ),
    put=extend_schema(
        operation_id="UpdateEntityType",
        description=dedent("""
        Update an existing entity type.

        This requires admin access to the entity type's corpus.
        """),
    ),
    patch=extend_schema(
        operation_id="PartialUpdateEntityType",
        description=dedent("""
        Update parts of an existing entity type.

        This requires admin access to the entity type's corpus.
        """),
    ),
    delete=extend_schema(
        operation_id="DestroyEntityType",
        description=dedent("""
        Delete an existing entity type if it has no associated entities.

        This requires admin access to the entity type's corpus.
        """)
    )
)
@extend_schema(tags=["entities"])
class EntityTypeUpdate(ACLMixin, RetrieveUpdateDestroyAPIView):
    serializer_class = EntityTypeSerializer
    permission_classes = (IsVerifiedOrReadOnly, )

    def get_queryset(self):
        return EntityType.objects.filter(corpus__in=Corpus.objects.readable(self.request.user))

    def check_object_permissions(self, request, obj):
        super().check_object_permissions(request, obj)

        if request.method in permissions.SAFE_METHODS:
            return
        if not self.has_access(obj.corpus, Role.Admin.value):
            raise PermissionDenied(detail="You do not have admin access to this corpus.")

    def perform_destroy(self, instance):
        if instance.entities.exists():
            raise ValidationError({"detail": ["Some entities are using this entity type."]})
        if EntityRole.objects.filter(Q(parent_type_id=instance.id) | Q(child_type_id=instance.id)).exists():
            raise ValidationError({"detail": ["Some entity roles are using this entity type."]})
        super().perform_destroy(instance)


@extend_schema(tags=["entities"])
@extend_schema_view(
    patch=extend_schema(description="Partially update an entity"),
    put=extend_schema(description="Update an entity"),
    delete=extend_schema(description="Delete an entity")
)
class EntityDetails(ACLMixin, RetrieveUpdateDestroyAPIView):
    """
    Get all information about an entity
    """
    permission_classes = (IsVerifiedOrReadOnly, )
    serializer_class = EntitySerializer

    def get_queryset(self):
        return Entity.objects \
            .select_related("corpus", "type") \
            .filter(corpus__in=Corpus.objects.readable(self.request.user)) \
            .prefetch_related(
                "parents__role__parent_type",
                "parents__role__child_type",
                "children__role__parent_type",
                "children__role__child_type",
                "parents__child__type",
                "parents__parent__type",
                "children__parent__type",
                "children__child__type",
                "corpus",
            )

    def check_object_permissions(self, request, obj):
        super().check_object_permissions(request, obj)
        if self.request.method not in permissions.SAFE_METHODS and not self.has_access(obj.corpus, Role.Contributor.value):
            raise PermissionDenied(detail="You do not have write access to this corpus.")


@extend_schema_view(get=extend_schema(operation_id="ListEntityElements", tags=["entities"]))
class EntityElements(ListAPIView):
    """
    Get all elements that have a link with the entity
    """
    serializer_class = ElementTinySerializer
    # For OpenAPI type discovery: an entity's ID is in the path
    queryset = Entity.objects.none()

    def get_queryset(self):
        pk = self.kwargs["pk"]
        metadata_elements = Element.objects \
            .filter(
                corpus__in=Corpus.objects.readable(self.request.user),
                metadatas__entity_id=pk
            ) \
            .select_related("type", "corpus") \
            .prefetch_related("metadatas__entity", "image__server")
        transcription_elements = Element.objects \
            .filter(
                corpus__in=Corpus.objects.readable(self.request.user),
                transcriptions__transcription_entities__entity_id=pk
            ).select_related("type", "corpus") \
            .prefetch_related("metadatas__entity", "image__server")
        return metadata_elements.union(transcription_elements) \
            .order_by("name", "type")


@extend_schema_view(post=extend_schema(
    operation_id="CreateEntity",
    responses={
        201: OpenApiResponse(
            response=EntitySerializer,
            description="When using the 'use_existing' parameter, if an entity with the same "
                        "name and type already exists in the corpus, this entity is returned.",
        ),
        200: EntitySerializer
    },
    tags=["entities"],
    parameters=[
        OpenApiParameter(
            "use_existing",
            type=bool,
            description="If an entity with the same name and type already exists in the corpus, "
                        "return this entity's UUID instead of creating a new one.",
            required=False,
        ),
    ]
))
class EntityCreate(CreateAPIView):
    """
    Create a new entity
    """
    permission_classes = (IsVerified, )
    serializer_class = EntityCreateSerializer

    def create(self, request, *args, **kwargs):
        # Overriding create in order to return EntitySerializer, not EntityCreateSerializer
        serializer = self.get_serializer(data=request.data)
        serializer.is_valid(raise_exception=True)

        name = serializer.validated_data["name"]
        type = serializer.validated_data["type"]
        corpus = serializer.validated_data["corpus"]
        metas = serializer.validated_data.get("metas")
        worker_run = serializer.validated_data["worker_run"]
        worker_version_id = worker_run.version_id if worker_run else None

        # When using the "use_existing" option, we return a 200_OK instead of a 201_CREATED status code
        if request.data.get("use_existing"):
            entity, created = Entity.objects.get_or_create(name=name, corpus=corpus, type=type, defaults={
                "metas": metas,
                "worker_version_id": worker_version_id,
                "worker_run": worker_run
            })
            entity = EntitySerializer(entity)
            if created:
                status_code = status.HTTP_201_CREATED
            else:
                status_code = status.HTTP_200_OK
        else:
            entity = EntitySerializer(Entity.objects.create(
                name=name,
                type=type,
                corpus=corpus,
                metas=metas,
                worker_version_id=worker_version_id,
                worker_run=worker_run
            ))
            status_code = status.HTTP_201_CREATED
        headers = self.get_success_headers(serializer.data)
        return Response(entity.data, status=status_code, headers=headers)


@extend_schema_view(post=extend_schema(operation_id="CreateEntityLink", tags=["entities"]))
class EntityLinkCreate(CreateAPIView):
    """
    Create a new link between two entities with a role
    """
    permission_classes = (IsVerified, )
    serializer_class = EntityLinkCreateSerializer


@extend_schema_view(post=extend_schema(
    operation_id="CreateTranscriptionEntity",
    tags=["entities"],
    parameters=[
        OpenApiParameter(
            "id",
            type=UUID,
            location=OpenApiParameter.PATH,
            description="ID of the transcription to link an entity to.",
            required=True,
        )
    ]
))
class TranscriptionEntityCreate(CreateAPIView):
    """
    Link an existing Entity to a given transcription with its position.
    """
    permission_classes = (IsVerified, )
    serializer_class = TranscriptionEntityCreateSerializer
    # For OpenAPI type discovery: a transcription's ID is in the path
    queryset = Transcription.objects.none()

    def get_serializer_context(self):
        context = super().get_serializer_context()
        if not self.request:
            # OpenAPI cannot pass request context
            return context

        # We need to use the default database to avoid stale read
        # when assigning a newly created entity on on a newly created transcription
        tr = Transcription.objects.using("default").filter(
            id=self.kwargs.get("pk"),
            element__corpus__in=Corpus.objects.writable(self.request.user)
        ).select_related("element").only("id", "text", "element__corpus_id").first()
        return {**context, "transcription": tr}

    def perform_create(self, serializer):
        data = serializer.validated_data
        try:
            TranscriptionEntity(**data).clean()
        except DjangoValidationError as e:
            raise serializers.ValidationError(serializers.as_serializer_error(e))
        super().perform_create(serializer)


@extend_schema_view(
    get=extend_schema(
        operation_id="ListTranscriptionEntities",
        tags=["entities"],
        parameters=[
            OpenApiParameter(
                "worker_version",
                type=UUID_OR_FALSE,
                description="Only include transcription entities created by a specific worker version. "
                "If set to `False`, only include transcription entities created by humans.",
                required=False,
            ),
            OpenApiParameter(
                "worker_run",
                type=UUID_OR_FALSE,
                description="Only include transcription entities created by a specific worker run. "
                "If set to `False`, only include transcription entities created by no worker run.",
                required=False,
            ),
            OpenApiParameter(
                "entity_worker_version",
                type=UUID_OR_FALSE,
                description="Only include transcription entities whose entity was created by a specific worker version. "
                "If set to `False`, only include transcription entities whose entity was created by humans.",
                required=False,
            ),
            OpenApiParameter(
                "entity_worker_run",
                type=UUID_OR_FALSE,
                description="Only include transcription entities whose entity was created by a specific worker run. "
                "If set to `False`, only include transcription entities whose entity was created by no worker run.",
                required=False,
            ),
        ]
    )
)
class TranscriptionEntities(ListAPIView):
    """
    List all entities linked to a transcription, ordeded by their position on the text.

    A guest access is required on the corpus of the transcription's element.
    """
    serializer_class = TranscriptionEntitySerializer
    # For OpenAPI type discovery: a transcription's ID is in the path
    queryset = Transcription.objects.none()

    def parse_model_id(self, model, value):
        if value.lower() in ("false", "0"):
            return None

        try:
            validated = UUID(value)
        except (TypeError, ValueError):
            raise serializers.ValidationError(["Invalid UUID."])

        if not model.objects.filter(id=validated).exists():
            raise serializers.ValidationError([f"This {model._meta.verbose_name} does not exist."])

        return validated

    def get_queryset(self):
        filters = {}
        errors = defaultdict()

        # List of WorkerVersion and WorkerRun filters to handle:
        # (query parameter, QuerySet field name, model class)
        worker_filters = (
            ("worker_version", "worker_version_id", WorkerVersion),
            ("entity_worker_version", "entity__worker_version_id", WorkerVersion),
            ("worker_run", "worker_run_id", WorkerRun),
            ("entity_worker_run", "entity__worker_run_id", WorkerRun),
        )

        for query_param, field_name, model in worker_filters:
            if query_param in self.request.query_params:
                try:
                    value = self.parse_model_id(model, self.request.query_params[query_param])
                    filters[field_name] = value
                except serializers.ValidationError as e:
                    errors[query_param] = e.detail

        if errors:
            raise serializers.ValidationError(errors)

        transcription = get_object_or_404(
            Transcription.objects
            .using("default")
            .filter(id=self.kwargs["pk"], element__corpus__in=Corpus.objects.readable(self.request.user))
            .only("id")
        )

        return (
            TranscriptionEntity.objects
            .filter(
                transcription=transcription,
                **filters,
            )
            .order_by("offset")
            .select_related("entity__type", "worker_run")
        )


@extend_schema_view(
    get=extend_schema(
        operation_id="ListCorpusEntities",
        tags=["entities"],
        parameters=[
            OpenApiParameter(
                "parent",
                type=UUID,
                description="Restrict entities to those linked to all transcriptions of an element "
                            "and all its descendants. Note that links to metadata are ignored.",
                required=False,
            ),
            OpenApiParameter(
                "name",
                type=str,
                description="Filter entities by part of their name (case-insensitive)",
                required=False,
            )
        ]
    )
)
class CorpusEntities(CorpusACLMixin, ListAPIView):
    """
    List all entities in a corpus.
    """
    serializer_class = BaseEntitySerializer

    def get_queryset(self):
        corpus = self.get_corpus(self.kwargs["pk"])
        queryset = corpus.entities.order_by("name", "id").select_related("type")

        if self.request.query_params.get("parent"):
            try:
                parent_id = UUID(self.request.query_params.get("parent"))
            except (TypeError, ValueError):
                raise serializers.ValidationError({"parent": ["Parent ID is not a valid UUID"]})

            parent_query = corpus.elements.filter(id=parent_id)
            if not parent_query.exists():
                raise NotFound

            # We want to include entities that are either related to the selected element, or to its descendants
            # Element.objects.get_descending() does not include the parent element, so we add it manually using a union.
            # We remove the ordering from get_descending as it would add an extra column to the query and cause
            # the union to fail.
            # Filtering using Q(element_id=…) | Q(element_id__in=…) is 100 times slower than filtering using a union.

            element_ids = parent_query.values("id").union(Element.objects.get_descending(parent_id).order_by())
            queryset = queryset.filter(transcriptions__element_id__in=element_ids)

        if self.request.query_params.get("name"):
            queryset = queryset.filter(name__icontains=self.request.query_params["name"])

        return queryset


@extend_schema_view(get=extend_schema(operation_id="ListElementLinks", tags=["entities"]))
class ElementLinks(CorpusACLMixin, ListAPIView):
    """
    List all links where parent and child are linked to the element.\n\n
    Requires a **guest** access to the element corpus
    """
    serializer_class = EntityLinkSerializer

    def get_queryset(self):
        try:
            element = Element.objects.select_related("corpus").only("id", "corpus").get(id=self.kwargs["pk"])
        except Element.DoesNotExist:
            raise NotFound

        if not self.has_read_access(element.corpus):
            raise PermissionDenied(detail="You do not have access to this element.")

        # Load entities linked by transcriptions
        entities_tr = Entity.objects.filter(transcriptions__element_id=element.id).prefetch_related("transcriptions")

        # Load entities linked by metadatas
        entities_meta = Entity.objects.filter(metadatas__element_id=element.id).prefetch_related("metadatas")

        # Now load all links belonging to those entities
        # It's several times faster to combine the queries in the final one
        # than combining them at the lower level (entities is slower than entities_tr + entities_meta)
        # We need to support cross references between transcriptions & metadata entities
        return EntityLink.objects.filter(
            Q(parent__in=entities_tr, child__in=entities_tr)
            | Q(parent__in=entities_tr, child__in=entities_meta)
            | Q(parent__in=entities_meta, child__in=entities_tr)
            | Q(parent__in=entities_meta, child__in=entities_meta)
        ).select_related("role", "child__type", "parent__type").order_by("parent__name")


@extend_schema_view(
    post=extend_schema(
        operation_id="CreateTranscriptionEntities",
        tags=["entities"],
    )
)
class TranscriptionEntitiesBulk(CorpusACLMixin, CreateAPIView):
    """
    Create multiple entities attached to a transcription.

    This requires the transcription to not already have any TranscriptionEntities from the specified WorkerRun.

    Requires a **contributor** access to the transcription's corpus.
    """
    serializer_class = TranscriptionEntitiesBulkSerializer
    permission_classes = (IsVerified, )
    # We need to use the default database to avoid stale read on a created transcription
    queryset = Transcription.objects.select_related("element__corpus").using("default")

    def get_object(self):
        transcription = super().get_object()
        if not self.has_write_access(transcription.element.corpus):
            raise PermissionDenied(detail="You do not have a contributor access to the corpus of this transcription.")
        return transcription

    def get_serializer_context(self):
        return {
            **super().get_serializer_context(),
            "transcription": self.get_object(),
        }
